/*
 *
 * ATM Switch Control Library - swctl
 * 	by miwa@mt.cs.keio.ac.jp, 1997
 * $Id: swctl.c,v 1.1.4.2 1998/09/14 07:08:09 miwa Exp $
 *
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ucd-snmp/ucd-snmp-config.h>
#include <ucd-snmp/ucd-snmp-includes.h>

#include "config.h"

#include "swctl.h"
#include "swctl_snmp.h"

/* GLOVAL VARIABLES */
static	swctl_switch_attr_t *sw_attr_list = (swctl_switch_attr_t *)NULL;
static	char SWCTL_VERSION[] = "swctl_version_0.2.5";

/*			*
 *	swctl_open()	*
 *			*/
int
swctl_open(switch_id_t sw,
	   u_char *set_community,
	   u_char *get_community,
	   swctl_vcc_table_t **vcc_table_list_p)
{
    struct snmp_session _session;	/* snmp session structure */
    struct snmp_session *session = &_session;
    struct in_addr sw_addr;		/* switch ip address */
    swctl_switch_attr_t	*sw_attr;	/* switch attribute infomation */

    swctl_upc_table_t *upc_table;
    swctl_vcc_table_t *vcc_table;

    swctl_mib_object_t *mib_object_list = (swctl_mib_object_t *)NULL;
    swctl_mib_object_t *mib_object;

    int snmp_stat;			/* snmp return status */

#ifdef	DEBUG
    printf("swctl_open()\n");
#endif	/* DEBUG */    

    *(vcc_table_list_p) = (swctl_vcc_table_t *)NULL;

    sw_attr = switch_attr_search(sw_attr_list, sw);

    if(sw_attr)				/* ALREADY open */
      return(SWCTL_ENTRY_FAILURE);

    /* construct swctl_swich_attr_t */
    sw_attr = (swctl_switch_attr_t *)malloc(sizeof(swctl_switch_attr_t));
    memset(sw_attr, 0, sizeof(swctl_switch_attr_t));
    sw_attr->next = sw_attr_list;
    sw_attr->id = sw;
    if( set_community != (u_char *)NULL
       && strlen(set_community) != 0) {
	sw_attr->set_community_len = strlen(set_community);
	sw_attr->set_community = strdup(set_community);
    }
    else {
	sw_attr->set_community_len
	  = strlen((char *)DEFAULT_SNMP_SET_COMMUNITY);
	sw_attr->set_community = strdup(DEFAULT_SNMP_SET_COMMUNITY);
    }

    if( get_community != (u_char *)NULL
       && strlen(get_community) != 0) {
	sw_attr->get_community_len = strlen(get_community);
	sw_attr->get_community = strdup(get_community);
    }
    else {
	sw_attr->get_community_len
	  = strlen((char *)DEFAULT_SNMP_GET_COMMUNITY);
	sw_attr->get_community = strdup(DEFAULT_SNMP_GET_COMMUNITY);
    }

    sw_attr->upc_table_list = (swctl_upc_table_t *)NULL;
    sw_attr_list = sw_attr;

    /* construct struct snmp_session */
    memset(session, 0, sizeof(struct snmp_session));

    sw_addr.s_addr = sw;
    session->peername = inet_ntoa(sw_addr);
    session->version = SNMP_VERSION_1;
    session->community = sw_attr->set_community;
    session->community_len = sw_attr->set_community_len;
    session->retries = SNMP_DEFAULT_RETRIES;
    session->timeout = 2000000L;
    session->authenticator = NULL;

    /* clean junk upc */
    if( (snmp_stat
	 = swctl_snmp_bulkwalk(session, SNMP_UPC_STAT_PREFIX,
			       SNMP_UPC_PREFIX_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {
	free_mib_object_list(mib_object_list);
	goto bad;
    }

    clean_junk_object(session, mib_object_list);
    free_mib_object_list(mib_object_list);
    /* __clean junk upc__ */

    /* costruct upc_table_list */
    if( (snmp_stat
	 = swctl_snmp_bulkwalk(session, SNMP_UPC_ROOT_PREFIX,
			       SNMP_UPC_ROOT_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {

	free_mib_object_list(mib_object_list);
	goto bad;
    }	

    for( mib_object = mib_object_list; mib_object;
	mib_object = mib_object->next_object) {

	if( object_prefix_cmp(mib_object->name, mib_object->name_length,
			      SNMP_UPC_KEY_PREFIX,
			      SNMP_UPC_PREFIX_LENGTH) ) {

	    upc_table = (swctl_upc_table_t
			 *)malloc(sizeof(swctl_upc_table_t));
	    memset((void *)upc_table, 0, sizeof(swctl_upc_table_t));
	    upc_table->next_table = sw_attr->upc_table_list;
	    upc_table->index = mib_object->value;
	    upc_table->param = (swctl_param_t
				*)malloc(sizeof(swctl_param_t));
	    memset((void *)upc_table->param, 0, sizeof(swctl_param_t));
	    sw_attr->upc_table_list = upc_table;
	}
	
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_STAT_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +mib_object->name_length-1));
	    if(upc_table)
	      upc_table->status = mib_object->value;
	}

	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_PCR01_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->pcr01 = mib_object->value;
	}
	
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_SCR01_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->scr01 = mib_object->value;
	}
	       
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_MBS01_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +mib_object->name_length-1));

	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->mbs01 = mib_object->value;
	}
	
	
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_PCR0_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->pcr0 = mib_object->value;
	}
	
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_SCR0_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +SNMP_UPC_PREFIX_LENGTH));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->scr0 = mib_object->value;
	}
	       
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_MBS0_PREFIX,	
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->mbs0 = mib_object->value;
	}
		
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_CDVT_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->cdvt = mib_object->value;
	}
    
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_ACT_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->act = mib_object->value;
	}
	  
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_AAL5EPD_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->aal5epd = mib_object->value;
	}
    }
	
    free_mib_object_list(mib_object_list);
    /* __construct upc_table_list__ */

#ifdef	DEBUG
    print_upc_table_list(sw_attr->upc_table_list);
#endif	/* DEBUG */

    /* clean junk vcc */
    if( (snmp_stat =
	 swctl_snmp_bulkwalk(session, SNMP_VCC_STAT_PREFIX,
			     SNMP_VCC_PREFIX_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {

	free_mib_object_list(mib_object_list);
	goto bad;
    }
    
    clean_junk_object(session, mib_object_list);
    free_mib_object_list(mib_object_list);

    if( (snmp_stat =
	 swctl_snmp_bulkwalk(session, SNMP_VCC_STATUS_PREFIX,
			     SNMP_VCC_PREFIX_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {

	free_mib_object_list(mib_object_list);
	goto bad;
    }
    
    clean_junk_object(session, mib_object_list);
    free_mib_object_list(mib_object_list);
    /* __clean junk vcc__ */

    /* construct vcc_table_list */
    if( (snmp_stat =
	 swctl_snmp_bulkwalk(session, SNMP_VCC_PROTO_PREFIX,
			     SNMP_VCC_PREFIX_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {

	free_mib_object_list(mib_object_list);
	goto bad;
    }
    
    /* filtering SNMP_PROTO_PVC */
    for( mib_object = mib_object_list; mib_object;
	mib_object = mib_object->next_object ) {

	if( object_prefix_cmp(mib_object->name, mib_object->name_length,
			      SNMP_VCC_PROTO_PREFIX,
			      SNMP_VCC_PREFIX_LENGTH) ) {
            if( mib_object->value == SNMP_PROTO_PVC ) {

		vcc_table = (swctl_vcc_table_t
			     *)malloc(sizeof(swctl_vcc_table_t));
		memset(vcc_table, 0, sizeof(swctl_vcc_table_t));

		vcc_table->next_table = *(vcc_table_list_p);
		vcc_table->status = SNMP_VALID_VAL;	/* static */
		vcc_table->in 
		  = ATM2ADDR(*(mib_object->name+mib_object->name_length-6),
			     *(mib_object->name+mib_object->name_length-5),
			     *(mib_object->name+mib_object->name_length-4));
		vcc_table->out
		  = ATM2ADDR(*(mib_object->name+mib_object->name_length-3),
			     *(mib_object->name+mib_object->name_length-2),
			     *(mib_object->name+mib_object->name_length-1));
		vcc_table->param = (swctl_param_t *)NULL;
		*(vcc_table_list_p) = vcc_table;
	    }
	}
    }

    free_mib_object_list(mib_object_list);

    /* which upc param does the vcc use ? */
    if( (snmp_stat =
	 swctl_snmp_bulkwalk(session, SNMP_VCC_UPC_PREFIX,
			     SNMP_VCC_PREFIX_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {

	free_mib_object_list(mib_object_list);
	goto bad;
    }

    for( mib_object = mib_object_list; mib_object;
	mib_object = mib_object->next_object) {

	if( object_prefix_cmp(mib_object->name, mib_object->name_length,
			      SNMP_VCC_UPC_PREFIX,
			      SNMP_VCC_PREFIX_LENGTH)
	    && mib_object->value > 0 ) {
	    for(vcc_table = *(vcc_table_list_p); vcc_table;
		vcc_table = vcc_table->next_table) {
		if( ADDR2VPORT(vcc_table->in)
		    == *(mib_object->name+mib_object->name_length-3)
		    && ADDR2VPI(vcc_table->in)
		    == *(mib_object->name+mib_object->name_length-2)
		    && ADDR2VCI(vcc_table->in)
		    == *(mib_object->name+mib_object->name_length-1) ) {
		    vcc_table->upc_index = mib_object->value;
		    upc_table =
		  	upc_search_index2table(sw_attr->upc_table_list,
					 	mib_object->value);
		    if(upc_table) {
		    	vcc_table->param = (swctl_param_t
					    *)malloc(sizeof(swctl_param_t));
		    	memcpy((void *)(vcc_table->param),
			       (void *)(upc_table->param),
			       sizeof(swctl_param_t));
		    }
		}
	    }
	}
    }

    free_mib_object_list(mib_object_list);
    /* __construct vcc_table_list__ */

#ifdef	DEBUG
    print_vcc_table_list(*(vcc_table_list_p));
#endif	/* DEBUG */

    return(SWCTL_SUCCESS);

bad:
    sw_attr_list = switch_attr_delete(sw_attr_list, sw);

    switch(snmp_stat) {
      case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);
	
      case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);
	
      default:
	return(SWCTL_UNKNOWN_FAILURE);
    }
}


/*			*
 *	swctl_close()	*
 *			*/
int
swctl_close(switch_id_t sw)
{
    struct snmp_session _session;	/* snmp session structure */
    struct snmp_session *session = &_session;
    struct in_addr sw_addr;		/* switch ip address */
    swctl_switch_attr_t *sw_attr;	/* switch attribute infomation */

    swctl_upc_table_t	*upc_table;

    swctl_mib_object_t	*mib_object_list = (swctl_mib_object_t *)NULL;
    swctl_mib_object_t	*mib_object;

    int snmp_stat;	       		/* snmp return status */

#ifdef	DEBUG
    printf("swctl_close()\n");
#endif	/* DEBUG */    

    sw_attr = switch_attr_search(sw_attr_list, sw);

    if(!sw_attr)			/* NOT open */
      return(SWCTL_ENTRY_FAILURE);

    /* session structure constract */
    memset(session, 0, sizeof(struct snmp_session));

    sw_addr.s_addr = sw;
    session->peername = inet_ntoa(sw_addr);
    session->version = SNMP_VERSION_1;
    session->community = sw_attr->set_community;
    session->community_len = sw_attr->set_community_len;
    session->retries = SNMP_DEFAULT_RETRIES;
    session->timeout = 2000000L;
    session->authenticator = NULL;

    /* clean junk upc */
    if( (snmp_stat =
	 swctl_snmp_bulkwalk(session, SNMP_UPC_STAT_PREFIX,
			     SNMP_UPC_PREFIX_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {	

	free_mib_object_list(mib_object_list);
	goto bad;
    }

    clean_junk_object(session, mib_object_list);
    free_mib_object_list(mib_object_list);
    /* __clean junk upc__ */

    /* clean junk vcc */
    if( (snmp_stat =
	 swctl_snmp_bulkwalk(session, SNMP_VCC_STAT_PREFIX,
			     SNMP_VCC_PREFIX_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {

	free_mib_object_list(mib_object_list);
	goto bad;
    }

    clean_junk_object(session, mib_object_list);
    free_mib_object_list(mib_object_list);

    if( (snmp_stat =
	 swctl_snmp_bulkwalk(session, SNMP_VCC_STATUS_PREFIX,
			     SNMP_VCC_PREFIX_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {

	free_mib_object_list(mib_object_list);
	goto bad;
    }

    clean_junk_object(session, mib_object_list);
    free_mib_object_list(mib_object_list);
    /* __clean junk vcc__ */

    sw_attr_list = switch_attr_delete(sw_attr_list, sw);

    return(SWCTL_SUCCESS);

bad:
    sw_attr_list = switch_attr_delete(sw_attr_list, sw);

    switch(snmp_stat) {
      case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);

      case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);

      default:
	return(SWCTL_UNKNOWN_FAILURE);
    }
}


/*			*
 *	swctl_add()	*
 *			*/
int
swctl_add(switch_id_t sw, swctl_addr_t in, swctl_addr_t out,
	  swctl_param_t *param)
{
    struct snmp_session _session;	/* snmp session structure */
    struct snmp_session *session = &_session;
    struct in_addr sw_addr;		/* switch ip address */
    swctl_switch_attr_t *sw_attr;	/* switch attribute infomation */
    int snmp_stat;	       		/* snmp return status */
    int upc_index;			/* upc index */

    swctl_upc_table_t *upc_table;
    swctl_mib_object_t *vcc_object, *upc_object, *vcc_upc_object, *vcc_rt_object;
    swctl_mib_object_t *upc_object_list = (swctl_mib_object_t *)NULL;
    swctl_mib_object_t *return_object_list = (swctl_mib_object_t *)NULL;

#ifdef	DEBUG
    printf("swctl_add()\n");
#endif	/* DEBUG */

    sw_attr = switch_attr_search(sw_attr_list, sw);
    if(!sw_attr)			/* NOT open */
      return(SWCTL_ENTRY_FAILURE);

    /* session structure constract */
    memset(session, 0, sizeof(struct snmp_session));

    sw_addr.s_addr = sw;
    session->peername = inet_ntoa(sw_addr);
    session->version = SNMP_VERSION_1;
    session->community = sw_attr->set_community;
    session->community_len = sw_attr->set_community_len;
    session->retries = SNMP_DEFAULT_RETRIES;
    session->timeout = SNMP_DEFAULT_TIMEOUT;
    session->authenticator = NULL;

    /* vcc object */
    vcc_object = make_vcc_object(SNMP_VCC_STATUS_PREFIX,
				 SNMP_VCC_PREFIX_LENGTH,
				 in, SNMP_REQ_VAL);

    /* setup vcc (query) */
    if( (snmp_stat
	 = swctl_snmp_get(session, vcc_object,
			  (swctl_mib_object_t **)NULL))
       != SNMP_SUCCESS ) {

	/* setup vcc (request) */
	if( (snmp_stat
	     = swctl_snmp_set(session, vcc_object,
			      (swctl_mib_object_t **)NULL))
	   != SNMP_SUCCESS ) {

	    free_mib_object_list(vcc_object);
	    goto bad_vcc;
	}

	/* upc setup (if any) */
	if( param ) {

	    /* search upc index */
	    if(!upc_search_param2index(sw_attr->upc_table_list, param,
				       &upc_index)) {

		/* upc object */
		upc_object
		  = make_upc_object(SNMP_UPC_STAT_PREFIX,
				    SNMP_UPC_PREFIX_LENGTH,
				    upc_index, SNMP_REQ_VAL);
	
		/* upc contract (request) */
		if( (snmp_stat
		     = swctl_snmp_set(session, upc_object,
				      (swctl_mib_object_t **)NULL))
		   != SNMP_SUCCESS ) {

		    free_mib_object_list(upc_object);

		    vcc_object->value = SNMP_INVALID_VAL;
		    swctl_snmp_set(session, vcc_object,
				   (swctl_mib_object_t **)NULL);
		    free_mib_object_list(vcc_object);

		    goto bad_upc;
		}

		/* upc parameter object list */
		upc_object_list = make_upc_object_list(upc_index, param);

		/* upc parameter set */
		if( (snmp_stat
		     = swctl_snmp_set(session, upc_object_list,
				      (swctl_mib_object_t **)NULL))
		   != SNMP_SUCCESS ) {
		  
		    free_mib_object_list(upc_object_list);

		    vcc_object->value = SNMP_INVALID_VAL;
		    swctl_snmp_set(session, vcc_object,
				   (swctl_mib_object_t **)NULL);
		    free_mib_object_list(vcc_object);

		    upc_object->value = SNMP_INVALID_VAL;
		    swctl_snmp_set(session, upc_object,
				   (swctl_mib_object_t **)NULL);
		    free_mib_object_list(upc_object);

		    goto bad_upc;
		}
		free_mib_object_list(upc_object_list);

		/* upc contract (enabled) */
		upc_object->value = SNMP_VALID_VAL;

		if( (snmp_stat
		     = swctl_snmp_set(session, upc_object, 
				      (swctl_mib_object_t **)NULL) )
		   != SNMP_SUCCESS ) {

		    vcc_object->value = SNMP_INVALID_VAL;
		    swctl_snmp_set(session, vcc_object,
				   (swctl_mib_object_t **)NULL);
		    free_mib_object_list(vcc_object);

		    upc_object->value = SNMP_INVALID_VAL;
		    swctl_snmp_set(session, upc_object,
				   (swctl_mib_object_t **)NULL);
		    free_mib_object_list(upc_object);

		    goto bad_upc;
		}

		free_mib_object_list(upc_object);

		sw_attr->upc_table_list =
		  upc_index_entry(sw_attr->upc_table_list, upc_index, param);
	    }
	}
	else {
	    upc_index = 0;
	}
	
	/* upc index object */
	vcc_upc_object
	  = make_vcc_object(SNMP_VCC_UPC_PREFIX,
			    SNMP_VCC_PREFIX_LENGTH,
			    in, upc_index);  

	/* set upc index */
	if( (snmp_stat
	     = swctl_snmp_set(session, vcc_upc_object,
			      (swctl_mib_object_t **)NULL))
	   != SNMP_SUCCESS ) {

	    free_mib_object_list(vcc_upc_object);

	    vcc_object->value = SNMP_INVALID_VAL;
	    swctl_snmp_set(session, vcc_object, (swctl_mib_object_t **)NULL);
	    free_mib_object_list(vcc_object);

	    goto bad_upc;
	}

	free_mib_object_list(vcc_upc_object);

	/* setup vcc status (enable) */
	vcc_object->value = SNMP_VALID_VAL;

	if( (snmp_stat
	     = swctl_snmp_set(session, vcc_object,
			      (swctl_mib_object_t **)NULL))
	   != SNMP_SUCCESS ) {

	    vcc_object->value = SNMP_INVALID_VAL;
	    swctl_snmp_set(session, vcc_object, (swctl_mib_object_t **)NULL);
	    free_mib_object_list(vcc_object);

	    goto bad_upc;
	}
	free_mib_object_list(vcc_object);
    }

    /* vcc rt object */
    vcc_rt_object
      = make_vcc_rt_object(SNMP_VCC_STAT_PREFIX, SNMP_VCC_PREFIX_LENGTH,
			   in, out, SNMP_REQ_VAL);

    /* setup vcc rt (query) */
    if ( (snmp_stat = swctl_snmp_set(session, vcc_rt_object,
				     (swctl_mib_object_t **)NULL))
	!= SNMP_SUCCESS ) {
	
	free_mib_object_list(vcc_rt_object);

	goto bad_vcc;
    }

    /* setup vcc rt (establish) */
    vcc_rt_object->value = SNMP_VALID_VAL;

    if( (snmp_stat = swctl_snmp_set(session, vcc_rt_object,
				    (swctl_mib_object_t **)NULL))
       !=  SNMP_SUCCESS ) {	

	vcc_rt_object->value = SNMP_INVALID_VAL;
	swctl_snmp_set(session, vcc_rt_object, (swctl_mib_object_t **)NULL);
	free_mib_object_list(vcc_rt_object);

	goto bad_vcc;
    }

    free_mib_object_list(vcc_rt_object);

    swctl_trap(sw);			/* trap msg to swview */
    return(SWCTL_SUCCESS);

bad_vcc:

    switch(snmp_stat) {
	    
      case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);

      case SNMP_PARAM_FAILURE:
	return(SWCTL_VCC_FAILURE);

      case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);

      default:
	return(SWCTL_UNKNOWN_FAILURE);
    }

bad_upc:
    
    switch(snmp_stat) {

      case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);
	
      case SNMP_PARAM_FAILURE:
	return(SWCTL_UPC_FAILURE);
	
      case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);
	
      default:
	return(SWCTL_UNKNOWN_FAILURE);
    }

}


/*			*
 *	swctl_delete()	*
 *			*/

int
swctl_delete(switch_id_t sw, swctl_addr_t in, swctl_addr_t out)
{
    struct snmp_session _session;	/* snmp session structure */
    struct snmp_session *session = & _session;
    struct in_addr sw_addr;		/* switch ip address */
    swctl_switch_attr_t *sw_attr;	/* switch attribute infomation */
    int	snmp_stat;			/* snmp status value */

    int upc_index = 0;

    swctl_mib_object_t *vcc_rt_object, *upc_object, *vcc_upc_object;
    swctl_mib_object_t	*return_object_list = (swctl_mib_object_t *)NULL;

#ifdef	DEBUG
    printf("swctl_delete()\n");
#endif	/* DEBUG */

    sw_attr = switch_attr_search(sw_attr_list, sw);
    if(!sw_attr)
      return(SWCTL_ENTRY_FAILURE);

    memset(session, 0, sizeof(struct snmp_session));

    sw_addr.s_addr = sw;
    session->peername = inet_ntoa(sw_addr);
    session->version = SNMP_VERSION_1;
    session->community = sw_attr->set_community;
    session->community_len = sw_attr->set_community_len;
    session->retries = SNMP_DEFAULT_RETRIES;
    session->timeout = SNMP_DEFAULT_TIMEOUT;
    session->authenticator = NULL;

    /* upc index object */
    vcc_upc_object = make_vcc_object(SNMP_VCC_UPC_PREFIX,
				     SNMP_VCC_PREFIX_LENGTH,
				     in, 0);

    /* which upc does the vcc use ? */
    if ( (snmp_stat = swctl_snmp_get(session, vcc_upc_object, 
				     &return_object_list))
	!= SNMP_SUCCESS ) {

	free_mib_object_list(return_object_list);
	free_mib_object_list(vcc_upc_object);

	goto bad_vcc;
    }

    upc_index = return_object_list->value;

    free_mib_object_list(return_object_list);
    free_mib_object_list(vcc_upc_object);

    /* vcc rt object */
    vcc_rt_object
      = make_vcc_rt_object(SNMP_VCC_STAT_PREFIX, SNMP_VCC_PREFIX_LENGTH,
			   in, out, SNMP_INVALID_VAL);

    /* vcc release */
    if( (snmp_stat = swctl_snmp_set(session, vcc_rt_object,
				    (swctl_mib_object_t **)NULL))
       != SNMP_SUCCESS ) {

	free_mib_object_list(vcc_rt_object);

	goto bad_vcc;
    }
    free_mib_object_list(vcc_rt_object);
    
    if( upc_index > 0 ) {

	/* delete upc entry */
	upc_object = make_upc_object(SNMP_UPC_STAT_PREFIX,
				     SNMP_UPC_PREFIX_LENGTH,
				     upc_index, SNMP_INVALID_VAL);
	
	if( (snmp_stat = swctl_snmp_set(session, upc_object,
					(swctl_mib_object_t **)NULL))
	   != SNMP_SUCCESS )
	  ; /* just ignored */
	else
	  sw_attr->upc_table_list =
	    upc_index_delete(sw_attr->upc_table_list, upc_index);

	free_mib_object_list(upc_object);
    }

    swctl_trap(sw);			/* trap msg to swview */
    return(SWCTL_SUCCESS);

bad_vcc:

    switch(snmp_stat) {
	    
      case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);

      case SNMP_PARAM_FAILURE:
	return(SWCTL_VCC_FAILURE);

      case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);

      default:
	return(SWCTL_UNKNOWN_FAILURE);
    }
}


/*			*
 *	swctl_modify() 	*
 *			*/

int
swctl_modify(switch_id_t sw, swctl_addr_t in, swctl_addr_t out,
	     swctl_param_t *param)
{
    struct snmp_session _session;	/* snmp sesseion structure */
    struct snmp_session *session = &_session;
    struct in_addr sw_addr;		/* switch ip address */
    swctl_switch_attr_t	*sw_attr;	/* switch attribute infomation */

    swctl_upc_table_t	*upc_table;

    int	upc_index = 0;
    int snmp_stat;	       		/* snmp return status */

    swctl_mib_object_t	*vcc_object, *upc_object, *vcc_upc_object;
    swctl_mib_object_t	*upc_object_list = (swctl_mib_object_t *)NULL;
    swctl_mib_object_t	*return_object_list = (swctl_mib_object_t *)NULL;
    
#ifdef	DEBUG
    printf("swctl_modify()\n");
#endif	/* DEBUG */

    sw_attr = switch_attr_search(sw_attr_list, sw);

    if(!sw_attr)			/* NOT open */
      return(SWCTL_ENTRY_FAILURE);

    memset(session, 0, sizeof(struct snmp_session));

    sw_addr.s_addr = sw;
    session->peername = inet_ntoa(sw_addr);
    session->version = SNMP_VERSION_1;
    session->community = sw_attr->set_community;
    session->community_len = sw_attr->set_community_len;
    session->retries = SNMP_DEFAULT_RETRIES;
    session->timeout = SNMP_DEFAULT_TIMEOUT;
    session->authenticator = NULL;

    /* vcc object */
    vcc_object
      = make_vcc_rt_object(SNMP_VCC_STAT_PREFIX, SNMP_VCC_PREFIX_LENGTH,
			   in, out, 0);

    /* vcc query */
    if( (snmp_stat
	 = swctl_snmp_get(session, vcc_object, &return_object_list))
       != SNMP_SUCCESS )    {
	free_mib_object_list(return_object_list);
	free_mib_object_list(vcc_object);
	goto bad_vcc;
    }
    else if( return_object_list->value != SNMP_VALID_VAL ) {
	free_mib_object_list(return_object_list);

	vcc_object->value = SNMP_INVALID_VAL;
	swctl_snmp_set(session, vcc_object, (swctl_mib_object_t **)NULL);
	free_mib_object_list(vcc_object);
	return(SWCTL_VCC_FAILURE);
    }

    free_mib_object_list(return_object_list);
    free_mib_object_list(vcc_object);

    if( param ) {

	if(!upc_search_param2index(sw_attr->upc_table_list, param,
				   &upc_index)) {
	  
	    /* upc object */
	    upc_object
	      = make_upc_object(SNMP_UPC_STAT_PREFIX, SNMP_UPC_PREFIX_LENGTH, 
				upc_index, SNMP_REQ_VAL);

	    /* upc contract (request) */
	    if( (snmp_stat = swctl_snmp_set(session, upc_object,
					    (swctl_mib_object_t **)NULL))
	       != SNMP_SUCCESS ) {
		
		upc_object->value = SNMP_INVALID_VAL;
		swctl_snmp_set(session, upc_object,
			       (swctl_mib_object_t **)NULL);
		free_mib_object_list(upc_object);

		goto bad_upc;
	    }
	    
	    /* upc parameter object list */
	    upc_object_list = make_upc_object_list(upc_index, param);

	    /* upc parameter set */
	    if( (snmp_stat = swctl_snmp_set(session, upc_object_list,
					    (swctl_mib_object_t **)NULL))
	       != SNMP_SUCCESS ) {

		free_mib_object_list(upc_object_list);

		upc_object->value = SNMP_INVALID_VAL;
		swctl_snmp_set(session, upc_object,
			       (swctl_mib_object_t **)NULL);
		free_mib_object_list(upc_object);

		goto bad_upc;
	    }
	
	    free_mib_object_list(upc_object_list);

	    /* upc contract (enable) */
	    upc_object->value = SNMP_VALID_VAL;

	    if( (snmp_stat = swctl_snmp_set(session, upc_object,
					    (swctl_mib_object_t **)NULL))
	       != SNMP_SUCCESS ) {

		upc_object->value = SNMP_INVALID_VAL;
		swctl_snmp_set(session, upc_object,
			       (swctl_mib_object_t **)NULL);
		free_mib_object_list(upc_object);		

		goto bad_upc;
	    }
	       
	    free_mib_object_list(upc_object);
	    
	    sw_attr->upc_table_list =
	      upc_index_entry(sw_attr->upc_table_list, upc_index, param);
	}
    }
    else {	/* param is not defined */
	upc_index = 0;
    }

    /* upc index object */
    vcc_upc_object
      = make_vcc_object(SNMP_VCC_UPC_PREFIX, SNMP_VCC_PREFIX_LENGTH,
			in, upc_index);

    /* upc index set */
    if( (snmp_stat = swctl_snmp_set(session, vcc_upc_object, 
				    (swctl_mib_object_t **)NULL))
       != SNMP_SUCCESS ) {

	free_mib_object_list(vcc_upc_object);
	goto bad_upc;
    }
    free_mib_object_list(vcc_upc_object);

    swctl_trap(sw);			/* trap msg to swview */
    return(SWCTL_SUCCESS);

bad_vcc:
    switch(snmp_stat) {
	    
      case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);

      case SNMP_PARAM_FAILURE:	/* vcc is NOT established */
	return(SWCTL_VCC_FAILURE);

      case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);

      default:
	return(SWCTL_UNKNOWN_FAILURE);
    }

bad_upc:
    switch(snmp_stat) {
	    
	case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);

	case SNMP_PARAM_FAILURE:
	return(SWCTL_UPC_FAILURE);

	case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);

	default:
	return(SWCTL_UNKNOWN_FAILURE);
    }
}



/*				*
 *	swctl_redirect()	*
 *				*/
int
swctl_redirect(switch_id_t sw, swctl_addr_t in,
	       swctl_addr_t out, swctl_addr_t new_out)
{
    struct snmp_session _session;	/* snmp session structure */
    struct snmp_session *session = &_session;
    struct in_addr sw_addr;		/* switch ip address */
    swctl_switch_attr_t	*sw_attr;	/* switch attribute status */
    int snmp_stat;			/* snmp return status */

    swctl_mib_object_t	*vcc_rt_object;	

#ifdef	DEBUG
    printf("swctl_redirect()\n");
#endif	/* DEBUG */

    sw_attr = switch_attr_search(sw_attr_list, sw);

    if(!sw_attr)			/* NOT open */
      return(SWCTL_ENTRY_FAILURE);

    memset(session, 0, sizeof(struct snmp_session));

    sw_addr.s_addr = sw;
    session->peername = inet_ntoa(sw_addr);
    session->version = SNMP_VERSION_1;
    session->community = sw_attr->set_community;
    session->community_len = sw_attr->set_community_len;
    session->retries = SNMP_DEFAULT_RETRIES;
    session->timeout = SNMP_DEFAULT_TIMEOUT;
    session->authenticator = NULL;

    /* new vcc rt object */
    vcc_rt_object = make_vcc_rt_object(SNMP_VCC_STAT_PREFIX,
				       SNMP_VCC_PREFIX_LENGTH,
				       in, new_out, SNMP_REQ_VAL);

    /* vcc query */
    if ( (snmp_stat = swctl_snmp_set(session, vcc_rt_object, 
				     (swctl_mib_object_t **)NULL))
	!= SNMP_SUCCESS ) {
	free_mib_object_list(vcc_rt_object);
	goto bad_vcc;
    }	

    vcc_rt_object->value = SNMP_VALID_VAL;

    /* vcc establish */
    if( (snmp_stat = swctl_snmp_set(session, vcc_rt_object,
				    (swctl_mib_object_t **)NULL))
       !=  SNMP_SUCCESS ) {

	vcc_rt_object->value = SNMP_INVALID_VAL;
	swctl_snmp_set(session, vcc_rt_object, (swctl_mib_object_t **)NULL);
	free_mib_object_list(vcc_rt_object);
	goto bad_vcc;
    }
    free_mib_object_list(vcc_rt_object);

    /* old vcc object */
    vcc_rt_object = make_vcc_rt_object(SNMP_VCC_STAT_PREFIX,
				       SNMP_VCC_PREFIX_LENGTH,
				       in, out, SNMP_INVALID_VAL);

    /* old vcc release */
    if( (snmp_stat = swctl_snmp_set(session, vcc_rt_object,
				    (swctl_mib_object_t **)NULL))
       != SNMP_SUCCESS ) {
	/* do nothing */
    }
    free(vcc_rt_object);

    swctl_trap(sw);			/* trap msg to swview */
    return(SWCTL_SUCCESS);

bad_vcc:
    switch(snmp_stat) {
	    
      case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);

      case SNMP_PARAM_FAILURE:
	return(SWCTL_VCC_FAILURE);

      case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);

      default:
	return(SWCTL_UNKNOWN_FAILURE);
    }
}


/*				*
 *	swctl_get() (obsolite) 	*
 *				*/
int
swctl_get (switch_id_t sw, swctl_addr_t in, swctl_addr_t out,
	   swctl_param_t *param)
{
    struct snmp_session _session;	/* snmp session structure */
    struct snmp_session *session = &_session;
    struct in_addr sw_addr;		/* switch IP address */
    swctl_switch_attr_t	*sw_attr;	/* switch attribute infomation */
    int snmp_stat;	       		/* snmp return status */
    int upc_index = 0;

    swctl_upc_table_t	*upc_table;

    swctl_mib_object_t	*vcc_rt_object, *vcc_upc_object;
    swctl_mib_object_t	*return_object_list = (swctl_mib_object_t *)NULL;

#ifdef	DEBUG
    printf("swctl_get()\n");
#endif	/* DEBUG */

    memset((void *)param, 0, sizeof(swctl_param_t));

    sw_attr = switch_attr_search(sw_attr_list, sw);

    if(!sw_attr)			/* NOT open */
      return(SWCTL_ENTRY_FAILURE);

    memset(session, 0, sizeof(struct snmp_session));

    sw_addr.s_addr = sw;
    session->peername = inet_ntoa(sw_addr);
    session->version = SNMP_VERSION_1;
    session->community = sw_attr->set_community;
    session->community_len = sw_attr->set_community_len;
    session->retries = SNMP_DEFAULT_RETRIES;
    session->timeout = SNMP_DEFAULT_TIMEOUT;
    session->authenticator = NULL;

    /* vcc rt object */
    vcc_rt_object = make_vcc_rt_object(SNMP_VCC_STAT_PREFIX,
				       SNMP_VCC_PREFIX_LENGTH,
				       in, out, 0);

    /* vcc query */
    if ( (snmp_stat = swctl_snmp_get(session, vcc_rt_object,
				     &return_object_list))
	!= SNMP_SUCCESS ) {

	free_mib_object_list(return_object_list);
	free_mib_object_list(vcc_rt_object);
	goto bad_vcc;
    }
    else if ( return_object_list->value != SNMP_VALID_VAL ) {
	free_mib_object_list(return_object_list);

	vcc_rt_object->value = SNMP_INVALID_VAL;
	swctl_snmp_set(session, vcc_rt_object,
		       (swctl_mib_object_t **)NULL);
	free_mib_object_list(vcc_rt_object);
	return(SWCTL_VCC_FAILURE);
    }
	
    free_mib_object_list(return_object_list);
    free_mib_object_list(vcc_rt_object);
    
    /* upc index object */
    vcc_upc_object = make_vcc_object(SNMP_VCC_UPC_PREFIX,
				     SNMP_VCC_PREFIX_LENGTH,
				     in, 0);

    /* upc query */
    if ( (snmp_stat = swctl_snmp_get(session, vcc_upc_object, 
				     &return_object_list))
	!= SNMP_SUCCESS ) {

	free_mib_object_list(return_object_list);
	free_mib_object_list(vcc_upc_object);

	goto bad_vcc;
    }

    upc_index = return_object_list->value;

    free_mib_object_list(return_object_list);
    free_mib_object_list(vcc_upc_object);

    if ( upc_index > 0 ) {
	upc_table = upc_search_index2table(sw_attr->upc_table_list,
					   upc_index);
	if(upc_table) {
	    memcpy((void *)param, (void *)(upc_table->param),
		   sizeof(swctl_param_t));
	}
    }
        
    return(SWCTL_SUCCESS);

bad_vcc:
    switch(snmp_stat) {

      case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);

      case SNMP_PARAM_FAILURE:
	return(SWCTL_VCC_FAILURE);

      case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);

      default:
	return(SWCTL_UNKNOWN_FAILURE);
    }
}


/*			*
 *	swctl_mon()	*
 *			*/
int swctl_mon(switch_id_t sw, swctl_addr_t in, swctl_addr_t out,
	      u_int *cells)
{
    struct snmp_session _session;	/* snmp sessio structure */
    struct snmp_session *session = &_session;
    struct in_addr sw_addr;		/* switch IP address */
    swctl_switch_attr_t	*sw_attr;	/* switch attribute infomation */

    int snmp_stat;	       		/* snmp return status */

    swctl_mib_object_t	*vcc_mon_object = (swctl_mib_object_t *)NULL;
    swctl_mib_object_t	*mib_object;
    swctl_mib_object_t	*return_object_list = (swctl_mib_object_t *)NULL;

#ifdef	DEBUG
    printf("swctl_mon()\n");
#endif	/* DEBUG */

    sw_attr = switch_attr_search(sw_attr_list, sw);

    if(!sw_attr)			/* NOT open */
      return(SWCTL_ENTRY_FAILURE);

    memset(session, 0, sizeof(struct snmp_session));

    sw_addr.s_addr = sw;
    session->peername = inet_ntoa(sw_addr);
    session->version = SNMP_VERSION_1;
    session->community = sw_attr->set_community;
    session->community_len = sw_attr->set_community_len;
    session->retries = SNMP_DEFAULT_RETRIES;
    session->timeout = SNMP_DEFAULT_TIMEOUT;
    session->authenticator = NULL;

    /* vcc mon object */
    mib_object = make_vcc_object(SNMP_VCC_REJECT_PREFIX,
				 SNMP_VCC_PREFIX_LENGTH,
				 in, 0);
    mib_object->next_object = vcc_mon_object;
    vcc_mon_object = mib_object;

    mib_object = make_vcc_object(SNMP_VCC_CELL_PREFIX,
				 SNMP_VCC_PREFIX_LENGTH,
				 in, 0);
    mib_object->next_object = vcc_mon_object;
    vcc_mon_object = mib_object;

    /* vcc mon query */
    if ( (snmp_stat = swctl_snmp_get(session, vcc_mon_object,
				     &return_object_list))
	!= SNMP_SUCCESS ) {

	free_mib_object_list(return_object_list);
	free_mib_object_list(vcc_mon_object);
	goto bad_vcc;
    }

    *(cells) = return_object_list->value;

    free_mib_object_list(return_object_list);
    free_mib_object_list(vcc_mon_object);
        
    return(SWCTL_SUCCESS);

bad_vcc:
    switch(snmp_stat) {

      case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);

      case SNMP_PARAM_FAILURE:	/* vcc is NOT established */
	return(SWCTL_VCC_FAILURE);

      case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);

      default:
	return(SWCTL_UNKNOWN_FAILURE);
    }
}


/*			*
 *	swctl_stat()	*
 *			*/
int
swctl_stat(switch_id_t sw, swctl_vcc_table_t  **vcc_table_list_p,
	   swctl_upc_table_t **upc_table_list_p)
{
    struct snmp_session _session;	/* snmp session structure */
    struct snmp_session *session = &_session;
    struct in_addr sw_addr;		/* switch IP address */
    swctl_switch_attr_t	*sw_attr;	/* switch attribute infomation */
    int snmp_stat;			/* snmp return status */

    swctl_upc_table_t *upc_table, *upc_table_tmp;
    swctl_vcc_table_t *vcc_table;

    swctl_mib_object_t *mib_object_list = (swctl_mib_object_t *)NULL;
    swctl_mib_object_t *mib_object;
    *(vcc_table_list_p) = (swctl_vcc_table_t *)NULL;
    *(upc_table_list_p) = (swctl_upc_table_t *)NULL;

#ifdef	DEBUG
    printf("swctl_stat()\n");
#endif	/* DEBUG */

    sw_attr = switch_attr_search(sw_attr_list, sw);

    if(!sw_attr)			/* NOT open */
      return(SWCTL_ENTRY_FAILURE);

    /* construct struct snmp_session */
    memset(session, 0, sizeof(struct snmp_session));

    sw_addr.s_addr = sw;
    session->peername = inet_ntoa(sw_addr);
    session->version = SNMP_VERSION_1;
    session->community = sw_attr->set_community;
    session->community_len = sw_attr->set_community_len;
    session->retries = SNMP_DEFAULT_RETRIES;
    session->timeout = 2000000L;
    session->authenticator = NULL;

    /* costruct switch upc_table_list */
    if( (snmp_stat
	 = swctl_snmp_bulkwalk(session, SNMP_UPC_ROOT_PREFIX,
			       SNMP_UPC_ROOT_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {

	free_mib_object_list(mib_object_list);
	goto bad;
    }

    /* free old upc_table_list */
    free_upc_table_list(sw_attr->upc_table_list);
    sw_attr->upc_table_list = (swctl_upc_table_t *)NULL;

    /* constract switch upc_table_list */
    for( mib_object = mib_object_list; mib_object;
	mib_object = mib_object->next_object) {

	if( object_prefix_cmp(mib_object->name, mib_object->name_length,
			      SNMP_UPC_KEY_PREFIX,
			      SNMP_UPC_PREFIX_LENGTH) ) {

	    upc_table = (swctl_upc_table_t
			 *)malloc(sizeof(swctl_upc_table_t));
	    memset((void *)upc_table, 0, sizeof(swctl_upc_table_t));
	    upc_table->next_table = sw_attr->upc_table_list;
	    upc_table->index = mib_object->value;
	    upc_table->param = (swctl_param_t
				*)malloc(sizeof(swctl_param_t));
	    memset((void *)upc_table->param, 0, sizeof(swctl_param_t));
	    sw_attr->upc_table_list = upc_table;
	}

	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_STAT_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				      *(mib_object->name
					+mib_object->name_length-1));
	    if(upc_table) 
	      upc_table->status = mib_object->value;
	}

	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_PCR01_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					+mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->pcr01 = mib_object->value;
	}
	
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_SCR01_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					+mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->scr01 = mib_object->value;
	}
	       
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_MBS01_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					+mib_object->name_length-1));

	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->mbs01 = mib_object->value;
	}
	
	
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_PCR0_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					+mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->pcr0 = mib_object->value;
	}
	
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_SCR0_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					 +SNMP_UPC_PREFIX_LENGTH));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->scr0 = mib_object->value;
	}
	       
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_MBS0_PREFIX,	
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					+mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->mbs0 = mib_object->value;
	}
		
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_CDVT_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					+mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->cdvt = mib_object->value;
	}
    
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_ACT_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					+mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->act = mib_object->value;
	}
	  
	else if( object_prefix_cmp(mib_object->name, mib_object->name_length,
				   SNMP_UPC_AAL5EPD_PREFIX,
				   SNMP_UPC_PREFIX_LENGTH) ) {
	    upc_table
	      = upc_search_index2table(sw_attr->upc_table_list,
				       *(mib_object->name
					+mib_object->name_length-1));
	    if( upc_table && mib_object->value > 0 )
	      upc_table->param->aal5epd = mib_object->value;
	}
    }
	
    free_mib_object_list(mib_object_list);
    /* __construct switch upc_table_list__ */

    /* construct vcc_table_list */
    if( (snmp_stat
	 = swctl_snmp_bulkwalk(session, SNMP_VCC_ROOT_PREFIX,
			       SNMP_VCC_ROOT_LENGTH,
			       &mib_object_list))
       != SNMP_SUCCESS ) {

	free_mib_object_list(mib_object_list);
	goto bad;
    }

    /* filtering SNMP_PROTO_PVC */
    for( mib_object = mib_object_list; mib_object;
	mib_object = mib_object->next_object ) {

	if( object_prefix_cmp(mib_object->name, mib_object->name_length,
			      SNMP_VCC_PROTO_PREFIX,
			      SNMP_VCC_PREFIX_LENGTH) ) {
            if( mib_object->value == SNMP_PROTO_PVC ) {

		vcc_table = (swctl_vcc_table_t
			     *)malloc(sizeof(swctl_vcc_table_t));
		memset(vcc_table, 0, sizeof(swctl_vcc_table_t));

		vcc_table->next_table = *(vcc_table_list_p);
		vcc_table->status = SNMP_VALID_VAL;	/* static */
		vcc_table->in 
		  = ATM2ADDR(*(mib_object->name+mib_object->name_length-6),
			     *(mib_object->name+mib_object->name_length-5),
			     *(mib_object->name+mib_object->name_length-4));
		vcc_table->out
		  = ATM2ADDR(*(mib_object->name+mib_object->name_length-3),
			     *(mib_object->name+mib_object->name_length-2),
			     *(mib_object->name+mib_object->name_length-1));
		vcc_table->param = (swctl_param_t *)NULL;
		*(vcc_table_list_p) = vcc_table;
	    }
	}
    }

    free_mib_object_list(mib_object_list);

    /* which upc param does the vcc use ? */
    if( (snmp_stat =
	 swctl_snmp_bulkwalk(session, SNMP_VCC_UPC_PREFIX,
			     SNMP_VCC_PREFIX_LENGTH, &mib_object_list))
       != SNMP_SUCCESS ) {

	free_mib_object_list(mib_object_list);
	goto bad;
    }

    for( mib_object = mib_object_list; mib_object;
	mib_object = mib_object->next_object) {

	if( object_prefix_cmp(mib_object->name, mib_object->name_length,
			      SNMP_VCC_UPC_PREFIX,
			      SNMP_VCC_PREFIX_LENGTH)
	    && mib_object->value > 0 ) {
	    for(vcc_table = *(vcc_table_list_p); vcc_table;
		vcc_table = vcc_table->next_table) {
		if( ADDR2VPORT(vcc_table->in)
		    == *(mib_object->name+mib_object->name_length-3)
		    && ADDR2VPI(vcc_table->in)
		    == *(mib_object->name+mib_object->name_length-2)
		    && ADDR2VCI(vcc_table->in)
		    == *(mib_object->name+mib_object->name_length-1) ) {
		    vcc_table->upc_index = mib_object->value;
		    upc_table =
		  	upc_search_index2table(sw_attr->upc_table_list,
					 	mib_object->value);
		    if(upc_table) {
		    	vcc_table->param = (swctl_param_t
					    *)malloc(sizeof(swctl_param_t));
		    	memcpy((void *)(vcc_table->param),
			       (void *)(upc_table->param),
			       sizeof(swctl_param_t));
		    }
		}
	    }
	}
    }

    free_mib_object_list(mib_object_list);
    /* __construct vcc_table_list__ */

    /* construct return upc_table_list */
    for( upc_table = sw_attr->upc_table_list; upc_table;
	upc_table = upc_table->next_table ) {
	upc_table_tmp = (swctl_upc_table_t
			 *)malloc(sizeof(swctl_upc_table_t));
	upc_table_tmp->next_table = *(upc_table_list_p);
	upc_table_tmp->status = upc_table->status;
	upc_table_tmp->index = upc_table->index;
	upc_table_tmp->param = (swctl_param_t
				*)malloc(sizeof(swctl_param_t));
	if(upc_table->param)
	  memcpy((void *)(upc_table_tmp->param), (void *)(upc_table->param),
		 sizeof(swctl_param_t));
	*(upc_table_list_p) = upc_table_tmp;
    }
    /* __construct return upc_table_list__ */
    
    return(SWCTL_SUCCESS);

bad:
    switch(snmp_stat) {
      case SNMP_LOCAL_FAILURE:
	return(SWCTL_LOCAL_FAILURE);

      case SNMP_TIMEOUT:
	return(SWCTL_NETWORK_FAILURE);

      default:
	return(SWCTL_UNKNOWN_FAILURE);
    }
}
